import datetime
import json
import logging
import os
import yaml

from .. import exception, ssl
from . import base

logger = logging.getLogger(__name__)


class LocalFileTracking(base.Tracking):

    def __init__(self, ssl_blueprint_path, log_folder, **kwargs):
        r"""Track actions in a log file on local file system

        :param ssl_blueprint_path: local path to ssl blueprint
        :type ssl_blueprint_path: pathlib.Path
        :param \**kwargs: generic key/value parameters
        :type \**kwargs: dict
        """
        super(LocalFileTracking, self).__init__(ssl_blueprint_path=ssl_blueprint_path, **kwargs)

        # validate input blueprint
        self.ssl_blueprint = ssl.SslBlueprint(yaml_path=self.ssl_blueprint_path)

        # build log file path from blueprint name
        self.log_file_path = '{}.log'.format(os.path.join(os.path.expandvars(log_folder), self.ssl_blueprint.name))

    def __write_log(self, content, name=None, data_type=None):
        structured_content = {
            'timestamp': datetime.datetime.utcnow().strftime('%Y/%d/%m %I:%M:%S %p'),
            'name': name or 'log',
            'content': os.path.expandvars(content),
        }
        if data_type is not None:
            structured_content['data_type'] = data_type.value

        with open(self.log_file_path, 'a') as log_file:
            json.dump(structured_content, log_file, sort_keys=True)
            log_file.write('\n')

    def create(self, tracking_type, servers=None):
        """Create a tracking record with details of current SSL blueprint

        :param tracking_type: Type of tracking. Can be used to customized tracking record content.
        :type tracking_type: TrackingType
        :param servers: List of servers in scope of the action. All servers from config if None specified here.
        :type servers: list
        :return: Identifier for the created record
        :rtype: str
        """
        message = ''
        if tracking_type == base.TrackingType.Renewal:
            message = """Creation of the following certificate requested:

              - Certificate Authority (CA): {certificate_authority}
              - Server type: {server_type}
              - Certificate Type: {certificate_type}
              - Common_name: {common_name}
              - Subject Alternate Name: {sans}
             """.format(
                certificate_authority=self.ssl_blueprint.certificate.certificate_authority,
                server_type=', '.join(set(server['type'] for server in self.ssl_blueprint.servers)),
                certificate_type=self.ssl_blueprint.certificate.certificate_type,
                common_name=self.ssl_blueprint.certificate.common_name or '',
                sans=', '.join(self.ssl_blueprint.certificate.sans),
            )
        elif tracking_type == base.TrackingType.Synchronize:
            message = """Some servers have a missing or outdated SSL certificate for:
              - Certificate Authority (CA): {certificate_authority}
              - Certificate Type: {certificate_type}
              - Common_name: {common_name}
              - Subject Alternate Name: {sans}

            Valid certificate is already available and will be deployed on the following servers:{servers}
            """.format(
                certificate_authority=self.ssl_blueprint.certificate.certificate_authority,
                certificate_type=self.ssl_blueprint.certificate.certificate_type,
                common_name=self.ssl_blueprint.certificate.common_name or '',
                sans=', '.join(self.ssl_blueprint.certificate.sans),
                servers=yaml.dump(data=servers or [], indent=2, default_flow_style=False),
            )
        self.update(message=message)

    def save_data(self, name, data_type, local_path=None, content=None, **kwargs):
        r"""Save input data in tracking system

        :param name: name of the file to attach to the tracking record
        :type name: str
        :param data_type: type of data to save
        :type data_type: ssl.DataType
        :param local_path: local path to file to attach to the tracking record
        :type local_path: pathlib.Path
        :param content: content of the file to attach to the tracking record
        :type content: bytes
        :param \**kwargs: generic key/value parameters
        :type kwargs: dict
        """
        if local_path:
            content = local_path.read_bytes()
        self.__write_log(content=content.decode('utf-8'), name=name, data_type=data_type)

    def update(self, message):
        """Update tracking record

        :param message: text to add to tracking record
        :type message: str
        """
        self.__write_log(message)

    def refresh(self, record_id):
        """Update current tracking instance with last changes from tracking record on server side

        :param record_id: identifier of the record to refresh
        """
        logger.info("Nothing to do for 'refresh' LocalFileLogging implementation.")

    def retrieve_data(self, name=None, data_type=None, **kwargs):
        r"""Retrieve specified data from tracking system

        :param name: Name of file/data to retrieve
        :type name: str
        :param data_type: type of data to retrieve
        :type data_type: ssl.DataType
        :param \**kwargs: generic key/value parameters
        :type kwargs: dict
        :return: file content
        :rtype: byte
        """
        if name is None and data_type is None:
            raise ValueError("Either name or data_type must be provided, none given.")

        if os.path.isfile(self.log_file_path):
            with open(self.log_file_path, 'r') as log_file:
                # search data from the end to get most accurate one
                # this works fine as this log file is expected to be small enough to fit in memory
                for raw_line in reversed(log_file.readlines()):
                    json_line = json.loads(raw_line)
                    if (name and name == json_line.get('name')) or \
                            (data_type and data_type.value == json_line.get('data_type')):
                        return json_line['content'].encode('utf-8')
        error_message = 'No data found.'
        if name:
            error_message = "No data found with name '{}'".format(name)
        elif data_type:
            error_message = "No data found for file of type '{}'".format(data_type.name)
        raise exception.NotFound(error_message)

    def close_for_failure(self, message):
        """Specify action is completed with a failed status

        :param message: custom message
        :type message: str
        """
        self.__write_log(message)

    def close_for_success(self, message):
        """Specify action is completed with a success status

        :param message: custom message
        :type message: str
        """
        self.__write_log(message)



from .. import ssl
from . import base


class LocalCa(base.CaManager):
    """Class implementing a certificate authority based on a private key retrieved from CA storage
    """
    def __init__(
            self,
            ca_config,
            staging=True,
            storage_api=None,
            ca_private_key=None,
            ca_certificate=None,
            certificate_validity_days=90,
            **kwargs
    ):
        super(LocalCa, self).__init__(ca_config, staging=staging, storage_api=storage_api)
        self.certificate_validity_days = certificate_validity_days
        self.ca_private_key = self.storage_api.retrieve_data(
            name=ca_private_key,
            data_type=ssl.DataType.PrivateKey,
        )
        self.ca_certificate = self.storage_api.retrieve_data(
            name=ca_certificate,
            data_type=ssl.DataType.Certificate,
        )

    def get_signed_certificate(self, ssl_blueprint=None, csr_path=None, servers_api=None):
        return ssl.sign(
            csr=csr_path.read_bytes(),
            ca_key=self.ca_private_key,
            ca_cert=self.ca_certificate,
            validity_days=self.certificate_validity_days,
        )

    @property
    def is_automated_renewal_supported(self):
        return True



import tempfile

import pytest

from autossl import util
from autossl.server.local import LocalServer


@pytest.mark.parametrize('class_path, exception_type', [
    ('MyClass', ValueError),
    ('my.package.does.not.exist.MyClass', ImportError),
    ('autossl.server.local.Dummy', AttributeError),
])
def test_str_to_class_error(class_path, exception_type):
    with pytest.raises(exception_type):
        util.str_to_class(class_path=class_path)


@pytest.mark.parametrize('class_path,class_type', [
    ('autossl.server.local.LocalServer', LocalServer),
])
def test_str_to_class(class_path, class_type):
    assert util.str_to_class(class_path=class_path) == class_type


@pytest.mark.parametrize('path', [
    None,
    util.Path(tempfile.mkdtemp())
])
def test_TempDir(path):
    with util.TempDir(path=path) as temp_folder:
        assert temp_folder.path.exists()
    assert not temp_folder.path.exists()



import pytest

from autossl import ssl

from tests import util as tests_util


@pytest.fixture(scope='function')
def ssl_blueprint():
    yield ssl.SslBlueprint(
        yaml_path=tests_util.DATA_PATH / 'tst.ov.example.com_minimal.yaml',
        global_config_path=tests_util.DATA_PATH / 'global_config.yaml',
    )


def test_ca_config(ssl_blueprint):
    assert ssl_blueprint.ca_config.is_acme_supported()
    assert ssl_blueprint.ca_config.get_supported_certificate_types() == ['DV']
    assert ssl_blueprint.ca_config.get_acme_api(staging=True) == 'https://acme-staging-v02.api.letsencrypt.org'
    assert ssl_blueprint.ca_config.get_acme_api(staging=False) == 'https://acme-v02.api.letsencrypt.org'
    assert ssl_blueprint.ca_config.get_storage_config() == {
        'type': 'autossl.storage.local.LocalFileStorage',
        'name': 'lets_encrypt_account_key',
        'parameters': {
            'path': '/etc/keys'
        },
    }
    assert len(ssl_blueprint.ca_config.get_chain_of_trust()) == 2



from .__version__ import __title__, __description__, __url__, __version__  # noqa: F401
from .__version__ import __author__, __author_email__, __license__, __copyright__  # noqa: F401

from . import server, storage, tracking, ca_manager  # noqa: F401



# functions to help testing and made available for use in autossl plugins

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from cryptography import x509
import datetime
import random
import string

from autossl import ssl, util

DATA_PATH = util.Path(__file__).parent.joinpath('data')


def create_ca_certificate(ca_name, key_size=4096, certificate_validity_days=365):
    key = rsa.generate_private_key(public_exponent=65537, key_size=key_size, backend=default_backend())
    key_id = x509.SubjectKeyIdentifier.from_public_key(key.public_key())

    subject = issuer = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, u"%s" % ca_name)])

    now = datetime.datetime.utcnow()
    serial = x509.random_serial_number()
    cert = x509.CertificateBuilder() \
        .subject_name(subject) \
        .issuer_name(issuer) \
        .public_key(key.public_key()) \
        .serial_number(serial) \
        .not_valid_before(now) \
        .not_valid_after(now + datetime.timedelta(days=certificate_validity_days)) \
        .add_extension(key_id, critical=False) \
        .add_extension(x509.AuthorityKeyIdentifier(key_id.digest,
                                                   [x509.DirectoryName(issuer)],
                                                   serial),
                       critical=False) \
        .add_extension(x509.BasicConstraints(ca=True, path_length=3), critical=True) \
        .add_extension(x509.KeyUsage(digital_signature=True,
                                     content_commitment=False,
                                     key_encipherment=False,
                                     data_encipherment=False,
                                     key_agreement=False,
                                     key_cert_sign=True,
                                     crl_sign=True,
                                     encipher_only=False,
                                     decipher_only=False),
                       critical=True) \
        .sign(key, hashes.SHA256(), default_backend())

    cert = cert.public_bytes(serialization.Encoding.PEM)
    key = key.private_bytes(encoding=serialization.Encoding.PEM,
                            format=serialization.PrivateFormat.TraditionalOpenSSL,
                            encryption_algorithm=serialization.NoEncryption())
    return key, cert


def create_signed_certificate(csr_path, ca_crt_path=None, ca_key_path=None, certificate_validity_days=365):
    """Generate SSL certificate signed by Autossl CA
    :param csr_path: local path to CSR
    :type csr_path: pathlib.Path
    :param ca_crt_path: local path to CA certificate
    :type ca_crt_path: pathlib.Path
    :param ca_key_path: local path to CA private key
    :type ca_key_path: pathlib.Path
    :param certificate_validity_days: validity duration of certificate in days
    :type certificate_validity_days: int
    """
    # no CA specified => generate self signed cert from pkey generated on the fly
    if ca_crt_path is None or ca_key_path is None:
        ca_key, ca_crt = create_ca_certificate('autossl.com')
    else:
        ca_key = ca_key_path.read_bytes()
        ca_crt = ca_crt_path.read_bytes()
    csr = csr_path.read_bytes()
    return ssl.sign(csr, ca_key, ca_crt, certificate_validity_days)


def create_self_signed_certificate(crt_name, output_path, common_name, sans=None, certificate_validity_days=365):
    """Generate self signed SSL certificate
    :param crt_name: name of file generated (without extension)
    :type crt_name: str
    :param output_path: ouput directory to generate certificate
    :type output_path: pathlib.Path or str
    :param common_name: certificate common name
    :type common_name: str
    :param sans: certificate alternate names
    :type sans: list
    :param certificate_validity_days: validity duration of certificate in days
    :type certificate_validity_days: int
    """
    if not isinstance(output_path, util.Path):
        output_path = util.Path(str(output_path))
    key_path = output_path.joinpath(crt_name + '.key')
    key_content, csr_path = ssl.generate_csr(name=crt_name, common_name=common_name, sans=sans, output_path=output_path)
    crt_path = output_path.joinpath(crt_name + '.crt')
    crt_content = create_signed_certificate(
        csr_path=csr_path,
        certificate_validity_days=certificate_validity_days,
    )
    crt_path.write_bytes(crt_content)
    key_path.write_bytes(key_content)

    return key_path, crt_path


def create_autossl_signed_cert(ssl_blueprint, output_folder):
    """Create certificate signed by Autossl CA for specified blueprint and in specified location

    :param ssl_blueprint: SSL blueprint for which to generate certificate
    :type ssl_blueprint: ssl.SslBlueprint
    :param output_folder: local folder where to generate new key and signed certificate
    :type output_folder: pathlib.Path
    :return: 2-tuple(local path to private key, local path to signed certificate)
    :rtype: tuple(pathlib.Path, pathlib.Path)
    """
    if not isinstance(output_folder, util.Path):
        output_folder = util.Path(output_folder)

    ca_crt_path = DATA_PATH / 'ca' / 'autossl_ca.crt'
    ca_key_path = DATA_PATH / 'ca' / 'autossl_ca.key'

    crt_path = output_folder.joinpath(ssl_blueprint.name + '.crt')
    key_path = output_folder.joinpath(ssl_blueprint.name + '.key')

    # generate new CSR
    key_content, csr_path = ssl.generate_csr(
        name=ssl_blueprint.name,
        common_name=ssl_blueprint.certificate.common_name,
        sans=ssl_blueprint.certificate.sans,
        output_path=output_folder,
    )
    key_path.write_bytes(key_content)

    # sign a new certificate with the CA
    crt_content = create_signed_certificate(
        csr_path=csr_path,
        ca_crt_path=ca_crt_path,
        ca_key_path=ca_key_path,
    )
    crt_path.write_bytes(crt_content)

    return key_path, crt_path


def get_random_text(size=200):
    """
    :return: random printable characters, bytes encoded
    :rtype: byte
    """
    return ''.join(random.choice(string.printable) for _ in range(size)).encode('utf-8')


def get_random_ascii_string(size=20):
    return ''.join(random.choice(string.ascii_letters + '_') for _ in range(size))



import datetime

from cryptography.hazmat.backends import default_backend
from cryptography.x509.oid import NameOID, ExtensionOID
from cryptography import x509
import pytest

from autossl import ssl
from tests import util as tests_util


@pytest.mark.parametrize('common_name,sans',
                         [
                             ('example.com', []),
                             ('example.com', ['example.com', 'www.example.com', 'test.example.com']),
                             (None, ['example.com', 'www.example.com', 'test.example.com']),
                         ]
                         )
def test_generate_dv_csr_with_key_creation(tmp_path, common_name, sans):
    name = 'auto_example.com'
    key_content, csr_path = ssl.generate_csr(
        name=name,
        common_name=common_name,
        sans=sans,
        key_content=None,
        output_path=str(tmp_path),
    )

    # check key + csr files have been created
    assert key_content.startswith(b'-----BEGIN RSA PRIVATE KEY-----\n')
    assert csr_path.is_file()

    # check csr file has right name
    assert csr_path.name == '%s.csr' % name

    # load csr first
    csr = x509.load_pem_x509_csr(data=csr_path.read_bytes(), backend=default_backend())

    # check content of CSR
    if common_name is not None:
        assert len(csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)) == 1
        assert csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value == common_name
    else:
        # common name is optional
        assert len(csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)) == 0

    assert sans == [dns_name.value
                    for dns_name in csr.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value]


@pytest.mark.parametrize('common_name,sans',
                         [
                             ('example.com', []),
                             ('example.com', ['example.com', 'www.example.com', 'test.example.com']),
                             (None, ['example.com', 'www.example.com', 'test.example.com']),
                          ]
                         )
def test_generate_dv_csr_with_existing_key(tmp_path, common_name, sans):
    name = 'auto_example.com'
    existing_key_content = tests_util.DATA_PATH.joinpath('auto_example.com.key').read_bytes()
    key_content, csr_path = ssl.generate_csr(
        name=name,
        common_name=common_name,
        sans=sans,
        key_content=existing_key_content,
        output_path=str(tmp_path),
    )

    # check that returned key content is same than input key content
    assert key_content == existing_key_content

    # check csr files has been created
    assert csr_path.is_file()

    # check csr file has right name
    assert csr_path.name == '%s.csr' % name

    # load csr first
    csr = x509.load_pem_x509_csr(data=csr_path.read_bytes(), backend=default_backend())

    # check content of CSR
    if common_name is not None:
        assert len(csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)) == 1
        assert csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value == common_name
    else:
        # common name is optional
        assert len(csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)) == 0

    assert sans == [dns_name.value
                    for dns_name in csr.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value]


def test_generate_two_csr_with_same_key(tmp_path):
    name = 'auto_example.com'
    common_name = 'example.com'
    sans = ['example.com', 'www.example.com', 'test.example.com']
    key_content, csr_path = ssl.generate_csr(
        name=name,
        common_name=common_name,
        sans=sans,
        key_content=None,
        output_path=str(tmp_path),
    )
    # generate 2nd csr with same key
    key_content, csr2_path = ssl.generate_csr(
        name='test_' + name,
        common_name=common_name,
        sans=sans,
        key_content=key_content,
        output_path=str(tmp_path),
    )

    # csr should be exactly the same if same key and content are used
    assert csr_path.read_bytes() == csr2_path.read_bytes()


def test_generate_ov_csr(tmp_path):
    name = 'auto_example.com'
    common_name = 'example.com'
    sans = ['example.com', 'www.example.com', 'test.example.com']
    company_name = 'Autossl corporation'
    street_address = 'Newbury street'
    city = 'Boston'
    state = 'Massachusetts'
    postal_code = '02115'
    country_code = 'US'
    email_address = 'test@autossl.com'

    key_content, csr_path = ssl.generate_csr(
        name=name,
        common_name=common_name,
        company_name=company_name,
        street_address=street_address,
        city=city,
        state=state,
        postal_code=postal_code,
        country_code=country_code,
        email_address=email_address,
        sans=sans,
        key_content=None,
        output_path=str(tmp_path),
    )

    # check key + csr files have been created
    assert key_content.startswith(b'-----BEGIN RSA PRIVATE KEY-----\n')
    assert csr_path.is_file()

    # check csr file has right name
    assert csr_path.name == '%s.csr' % name

    # load csr first
    csr = x509.load_pem_x509_csr(data=csr_path.read_bytes(), backend=default_backend())

    # check content of CSR
    # common name
    assert len(csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)) == 1
    assert csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value == common_name
    # sans
    assert sans == [dns_name.value
                    for dns_name in csr.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value]
    # organization
    assert csr.subject.get_attributes_for_oid(NameOID.ORGANIZATION_NAME)[0].value == company_name
    assert csr.subject.get_attributes_for_oid(NameOID.STREET_ADDRESS)[0].value == street_address
    assert csr.subject.get_attributes_for_oid(NameOID.LOCALITY_NAME)[0].value == city
    assert csr.subject.get_attributes_for_oid(NameOID.STATE_OR_PROVINCE_NAME)[0].value == state
    assert csr.subject.get_attributes_for_oid(NameOID.POSTAL_CODE)[0].value == postal_code
    assert csr.subject.get_attributes_for_oid(NameOID.COUNTRY_NAME)[0].value == country_code
    assert csr.subject.get_attributes_for_oid(NameOID.EMAIL_ADDRESS)[0].value == email_address


@pytest.mark.parametrize('common_name,sans',
                         [
                             ('example.com', []),
                             ('example.com', ['example.com', 'www.example.com', 'test.example.com']),
                             (None, ['example.com', 'www.example.com', 'test.example.com']),
                          ]
                         )
def test_get_domains_from_x509(tmp_path, common_name, sans):
    name = 'auto_example.com'
    key_path, csr_path = ssl.generate_csr(
        name=name,
        common_name=common_name,
        sans=sans,
        key_content=None,
        output_path=str(tmp_path),
    )

    expected_domains = set(sans)
    if common_name is not None:
        expected_domains.add(common_name)

    assert ssl.get_domains_from_x509(csr_path, file_type=ssl.DataType.CertificateSigningRequest) == expected_domains


@pytest.mark.parametrize('common_name, sans, result', [
    ("test.autossl.com", ["san1.autossl.com", "san2.autossl.com"],
        {"test.autossl.com", "san1.autossl.com", "san2.autossl.com"}),
    (None, ["san1.autossl.com", "san2.autossl.com"], {"san1.autossl.com", "san2.autossl.com"}),
    ("test.autossl.com", None, {"test.autossl.com"}),
    ("test.autossl.com", [], {"test.autossl.com"}),
    ("test.autossl.com", ["test.autossl.com", "san1.autossl.com"], {"test.autossl.com", "san1.autossl.com"}),
    (None, ["san1.autossl.com", "san2.autossl.com"], {"san1.autossl.com", "san2.autossl.com"}),
])
def test_get_domains(common_name, sans, result):
    assert ssl.get_domains(common_name=common_name, sans=sans) == result


def test_get_expiration():
    assert ssl.get_expiration(crt_path=tests_util.DATA_PATH / 'test.crt') == datetime.datetime(2018, 2, 5, 20, 5, 22)


def test_check_certificate_with_key():
    # CRT and KEY are matching
    assert ssl.check_certificate_with_key(key_path=tests_util.DATA_PATH / 'test.key',
                                          crt_path=tests_util.DATA_PATH / 'test.crt') is True

    # CRT that does not match KEY
    assert ssl.check_certificate_with_key(key_path=tests_util.DATA_PATH / 'auto_example.com.key',
                                          crt_path=tests_util.DATA_PATH / 'test.crt') is False

    # invalid key path
    with pytest.raises(IOError):
        ssl.check_certificate_with_key(key_path=tests_util.DATA_PATH / 'dummy.key',
                                       crt_path=tests_util.DATA_PATH / 'test.crt')

    # invalid crt path
    with pytest.raises(IOError):
        ssl.check_certificate_with_key(key_path=tests_util.DATA_PATH / 'test.key',
                                       crt_path=tests_util.DATA_PATH / 'dummy.crt')


@pytest.mark.parametrize('common_name,sans,validity_days',
                         [
                            ('test.autossl.com', [], 60),
                            ('test.autossl.com', ['example.com', 'www.example.com', 'test.example.com'], 60),
                            ('example.com', [], 0),
                          ]
                         )
def test_cert_sign(tmp_path, common_name, sans, validity_days):
    ca_key, ca_crt = tests_util.create_ca_certificate(ca_name='Autossl')
    _, csr_path = ssl.generate_csr(
        name='autossl_cert',
        common_name=common_name,
        sans=sans,
        output_path=str(tmp_path),
    )

    signed_crt = ssl.sign(csr=csr_path.read_bytes(), ca_key=ca_key, ca_cert=ca_crt, validity_days=validity_days)

    # load CRT
    x509_object = x509.load_pem_x509_certificate(data=signed_crt, backend=default_backend())

    # check content of CRT
    assert len(x509_object.subject.get_attributes_for_oid(NameOID.COMMON_NAME)) == 1
    assert x509_object.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value == common_name
    assert sans == [dns_name.value for dns_name in
                    x509_object.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value]



import collections
import os
import shutil
import tempfile

import pytest

from autossl import __main__, ssl, util

from tests import util as tests_util

EnvConfig = collections.namedtuple('EnvConfig', 'server_path storage_path tracking_path')


def get_default_options(ssl_blueprint_path=None):
    """Convenient method for testing command line, that provides minimum necessary command line options"""
    return ['--staging', '--blueprint', str(ssl_blueprint_path)]


@pytest.fixture(scope='function')
def ssl_blueprint():
    """Provide bluepring instance using local apis"""
    yield ssl.SslBlueprint(yaml_path=tests_util.DATA_PATH / 'local.example.com.yaml')


@pytest.fixture(scope='function')
def env_config():
    # create temporary directories use by blueprint
    temp_crt_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_CRT_PATH'] = temp_crt_dir
    temp_storage_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_STORAGE_PATH'] = temp_storage_dir
    temp_tracking_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_TRACKING_PATH'] = temp_tracking_dir

    yield EnvConfig(util.Path(temp_crt_dir), util.Path(temp_storage_dir), util.Path(temp_tracking_dir))

    # cleanup generated artifacts
    shutil.rmtree(temp_crt_dir, ignore_errors=True)
    shutil.rmtree(temp_storage_dir, ignore_errors=True)


def test_main_version(capsys):
    __main__.main(['version'])
    assert capsys.readouterr().out.rstrip() == __main__.display_version()


def test_main_check_with_invalid_blueprint_path():
    with pytest.raises(IOError):
        __main__.main(get_default_options(ssl_blueprint_path='/does/not/exist') + ['check'])


def test_main_check_with_blueprint_file(env_config, ssl_blueprint):
    __main__.main(get_default_options(ssl_blueprint.blueprint_path) + ['check'])


def test_main_renew(env_config, ssl_blueprint):
    # ensure nothing generated yet
    assert len(os.listdir(str(env_config.server_path))) == 0
    assert len(os.listdir(str(env_config.server_path))) == 0
    assert len(os.listdir(str(env_config.storage_path))) == 0

    # request renewal
    __main__.main(get_default_options(ssl_blueprint.blueprint_path) + ['renew', '--force'])

    # no crt generated yet
    assert len(os.listdir(str(env_config.server_path))) == 0
    # but we should have stored new private key and csr
    assert len(os.listdir(str(env_config.storage_path))) == 2
    for file_ in env_config.storage_path.iterdir():
        assert file_.suffix in ['.key', '.csr']


def test_main_deploy_all_servers(env_config, ssl_blueprint):
    # generate valid key + associated crt
    tests_util.create_autossl_signed_cert(
        ssl_blueprint=ssl_blueprint,
        output_folder=env_config.storage_path,
    )

    # perform deployment
    __main__.main(get_default_options(ssl_blueprint.blueprint_path) + ['deploy', '--all-servers'])

    # crt and key should be deployed on server
    assert len(os.listdir(str(env_config.server_path))) == 2
    for file_ in os.listdir(str(env_config.server_path)):
        assert file_.split('.')[-1] in ['key', 'crt']

    # and we should have stored private key, csr and crt
    assert len(os.listdir(str(env_config.storage_path))) == 3
    for file_ in os.listdir(str(env_config.storage_path)):
        assert file_.split('.')[-1] in ['key', 'csr', 'crt']

    __main__.main(get_default_options(ssl_blueprint.blueprint_path) + ['check'])


def test_main_deploy_invalid_servers_only(env_config, ssl_blueprint):
    assert len(os.listdir(str(env_config.server_path))) == 0
    # generate valid key + associated crt
    tests_util.create_autossl_signed_cert(
        ssl_blueprint=ssl_blueprint,
        output_folder=env_config.storage_path,
    )

    __main__.main(get_default_options(ssl_blueprint.blueprint_path) + ['deploy'])

    assert len(os.listdir(str(env_config.server_path))) == 2



import logging

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from acme import challenges
from acme import client
from acme import errors
from acme import messages
import josepy as jose

from .. import ssl, util, exception
from . import base


USER_AGENT = 'python-autossl'

logger = logging.getLogger(__name__)


class AcmeHttp01(base.CaManager):

    def __init__(self, ca_config, staging=True, storage_api=None, **kwargs):
        super(AcmeHttp01, self).__init__(ca_config, staging=staging, storage_api=storage_api)
        self.ca_api = ca_config.get_acme_api(staging=self.staging)
        self.contact_email = ca_config.ca_config.get('contact_email')

        self.account_key = None
        self.servers_api = []

    def get_signed_certificate(self, ssl_blueprint=None, csr_path=None, servers_api=None):
        self.servers_api = servers_api

        with util.TempDir() as temp_folder:
            # retrieve CA account key from storage
            ca_account_key_content = self.storage_api.retrieve_data(
                name=self.ca_config.ca_config['storage']['name'],
                data_type=ssl.DataType.PrivateKey,
            )
            self.account_key = temp_folder.path / 'ca_account.key'
            self.account_key.write_bytes(ca_account_key_content)

            client_acme = self._get_or_create_account()

            # Request new certificate order for specified CSR
            order_resource = client_acme.new_order(csr_pem=csr_path.read_text())

            # signed certificate ready to be used
            pem_signed_certificate = self._perform_http01(client_acme, order_resource)

        # return PEM certificate (bytes)
        return pem_signed_certificate.encode('utf-8')

    @property
    def is_automated_renewal_supported(self):
        return True

    def _get_or_create_account(self):
        """Register account if not already done, else just return ACME client updated with existing account"""
        logger.info("Registering account...")

        ca_key = serialization.load_pem_private_key(
            data=self.account_key.read_bytes(),
            password=None,
            backend=default_backend(),
        )
        acc_key = jose.JWKRSA(key=ca_key)

        # Register account and accept TOS
        client_network = client.ClientNetwork(acc_key, user_agent=USER_AGENT)
        directory = messages.Directory.from_json(client_network.get(self.ca_api + '/directory').json())
        client_acme = client.ClientV2(directory=directory, net=client_network)

        try:
            # Creates account with contact information.
            # Terms of Service URL is in client_acme.directory.meta.terms_of_service
            client_acme.new_account(
                messages.NewRegistration.from_data(
                    email=self.contact_email,
                    terms_of_service_agreed=True,
                )
            )
            logger.info("New account successfully created!")
        except errors.ConflictError:
            logger.info("Account already registered!")
            # retrieve existing account information
            account_registration = messages.NewRegistration(key=acc_key.public_key(), only_return_existing=True)

            # TODO migrate to public API when available in ACME module
            #  reproduce code of client_acme.new_account but without throwing exception when account already exists
            response = client_acme._post(directory['newAccount'], account_registration)
            account_info = client_acme._regr_from_response(response)

            # assign registration information to existing client
            client_acme.net.account = account_info
        return client_acme

    @staticmethod
    def _get_http01_challenge(authz):
        """Extract HTTP01 challenge from authorization resource."""
        # authz.body.challenges is a set of ChallengeBody objects.
        for challenge_body_item in authz.body.challenges:
            # Find HTTP01 challenge for each domain to validate
            if isinstance(challenge_body_item.chall, challenges.HTTP01):
                return challenge_body_item
        raise exception.AutoSslException('HTTP-01 challenge was not offered by the CA server.')

    def _deploy_challenge(self, client_acme, authz):
        """Deploy challenge token on all servers and warn CA server that it can start validation"""
        challenge_body = self._get_http01_challenge(authz)
        response, validation = challenge_body.response_and_validation(client_acme.net.key)
        token = challenge_body.chall.encode('token')
        logger.info("Deploy challenge for domain {}".format(authz.body.identifier.value))
        try:
            # deploy challenge on servers
            for server_api in self.servers_api:
                server_api.create_acme_challenge(
                    token=token,
                    key_authorization=validation,
                )

            # Let the CA server know that we are ready for this challenge validation
            client_acme.answer_challenge(challenge_body, response)

        except Exception:
            # cleanup challenge in case of any exception, to avoid leaking tokens on servers
            self._cleanup_challenges(token=token)
            raise

        return token

    def _cleanup_challenges(self, token):
        """Remove specified challenge token from all servers"""
        for server_api in self.servers_api:
            server_api.delete_acme_challenge(token=token)

    def _perform_http01(self, client_acme, order_resource):
        """Set up challenge on servers and perform HTTP-01 challenge verification."""

        # keep track of all tokens successfully deployed to be able to clean them up at the end
        tokens = []
        try:
            # Deploy token/validation for HTTP-01 challenge
            for authz in order_resource.authorizations:
                tokens.append(self._deploy_challenge(client_acme, authz))

            # Wait for challenge status and then issue a certificate.
            finalized_order_resource = client_acme.poll_and_finalize(order_resource)
        finally:
            # cleanup all challenges from servers once all validation is done
            for token in tokens:
                self._cleanup_challenges(token=token)

        # return certificate
        return finalized_order_resource.fullchain_pem



__title__ = 'autossl'
__description__ = 'SSL certificates monitoring, renewal and deployment.'
__url__ = 'https://autossl.readthedocs.io'
__version__ = '0.9.0'
__author__ = 'Thibaud Castaing'
__author_email__ = 't-cas@users.noreply.github.com'
__license__ = 'MIT license'
__copyright__ = 'Copyright 2019 Amadeus S.A.S.'



import pytest

from autossl.ca_manager import base


@pytest.fixture(scope="module")
def ca_manager():
    yield base.CaManager(ca_config=None, staging=True, storage_api=None)


def test_automated_renewal_supported(ca_manager):
    assert ca_manager.is_automated_renewal_supported is False


def test_get_signed_certificate(ca_manager):
    with pytest.raises(NotImplementedError):
        ca_manager.get_signed_certificate(ssl_blueprint=None, csr_path=None, servers_api=None)



import os
from pkg_resources import parse_version
from sys import version_info as py_version

from setuptools import setup, find_packages
from setuptools import __version__ as setuptools_version

HERE = os.path.abspath(os.path.dirname(__file__))

# retrieve package information
about = {}
with open(os.path.join(HERE, 'autossl', '__version__.py'), 'rb') as f:
    exec(f.read().decode('utf-8'), about)

with open(os.path.join(HERE, 'README.rst'), 'rb') as readme_file:
    readme = readme_file.read().decode('utf-8')

install_requires = [
    'six',
    'cryptography',
    'pyyaml',
    'requests',
]

extras_require = {
    # acme
    'acme': ['acme'],
    # servers
    # tracking
    # storage
    'git': ['GitPython'],
}
# ability to install automatically all dependencies
extras_require['all'] = list(set(value for sublist in extras_require.values() for value in sublist))


def has_environment_marker_range_operators_support():
    """Code extracted from 'pytest/setup.py'
    https://github.com/pytest-dev/pytest/blob/7538680c/setup.py#L31

    The first known release to support environment marker with range operators
    it is 17.1, see: https://setuptools.readthedocs.io/en/latest/history.html#id113
    """
    return parse_version(setuptools_version) >= parse_version('17.1')


# Compatibility with old version of setuptools
if has_environment_marker_range_operators_support():
    extras_require[':python_version<"3.4"'] = ['enum34', 'pathlib2']
elif py_version < (3, 4):
    install_requires.extend(['enum34', 'pathlib2'])


setup(
    name=about['__title__'],
    version=about['__version__'],

    author=about['__author__'],
    author_email=about['__author_email__'],

    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    url=about['__url__'],
    license=about['__license__'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',

        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',

        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'autossl = autossl.__main__:main'
        ]
    },
    platforms='Unix; MacOS X',

    install_requires=install_requires,
    extras_require=extras_require,
)



import logging
import os
import socket

from .. import exception, ssl, util
from . import base

logger = logging.getLogger(__name__)


class LocalServer(base.Server):

    def __init__(self, crt_name, path, acme_dir=None, **kwargs):
        """
        :param crt_name:
        :type crt_name: str
        :param path:
        :type path: pathlib.Path or str
        :param acme_dir:
        :type acme_dir: pathlib.Path or str
        """
        super(LocalServer, self).__init__(crt_name=crt_name, **kwargs)

        if not isinstance(path, util.Path):
            path = util.Path(os.path.expandvars(str(path)))
        self.path = path

        if acme_dir and not isinstance(acme_dir, util.Path):
            acme_dir = util.Path(os.path.expandvars(str(acme_dir)))
        self.acme_dir = acme_dir

    def get_description(self):
        return "[{} - {}:{}]".format(self.__class__.__name__, socket.gethostname(), self.path)

    def deploy_cert(self, key, cert, **kwargs):
        try:
            for source_path, destination_name in [
                (key, self.crt_name + '.key'),
                (cert, self.crt_name + '.crt'),
            ]:
                destination_path = self.path / destination_name
                destination_path.write_bytes(source_path.read_bytes())

            logger.info("Certificate/Key {} updated successfully on {} in {}.".format(
                self.crt_name, self.get_description(), self.path))
        except IOError as ex:
            raise exception.DeployCertificateError(
                msg='Unable to deploy new certificate on {}:{}'.format(self.get_description(), str(ex)),
                original_exception=ex,
            )

    def create_acme_challenge(self, token, key_authorization):
        logger.info("Deploy challenge for token %s" % token)
        wellknown_path = self.acme_dir / token
        wellknown_path.write_bytes(key_authorization.encode('utf-8'))

    def delete_acme_challenge(self, token):
        wellknown_path = self.acme_dir / token
        if wellknown_path.is_file():
            logger.info("Cleanup challenge for token %s" % token)
            wellknown_path.unlink()

    def get_certificate_information(self):
        path = self.path / (self.crt_name + '.crt')
        # certificate is missing on server
        if not path.exists():
            raise exception.CertificateNotFound("Certificate missing on local server at {}".format(path))
        return ssl.SslCertificate(x509_path=path)



# standard packages
import os
import shutil
import tempfile
import collections

# external packages
import pytest

# autossl imports
from autossl import manager, ssl, util

from tests import util as tests_util

CertificateKeyPair = collections.namedtuple('CertificateKeyPair', 'key crt')


@pytest.fixture(scope='module')
def subca_manager():
    # create temporary directories use by blueprint
    temp_crt_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_CRT_PATH'] = temp_crt_dir
    temp_crt_dir_2 = tempfile.mkdtemp()
    os.environ['AUTOSSL_CRT_PATH_2'] = temp_crt_dir_2

    temp_storage_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_STORAGE_PATH'] = temp_storage_dir
    temp_tracking_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_TRACKING_PATH'] = temp_tracking_dir

    yield manager.SslManager(
        global_config=None,
        blueprint_path=tests_util.DATA_PATH / 'subca.example.com.yaml',
        credentials=None,
        staging=True
    )

    # cleanup generated artifacts
    shutil.rmtree(temp_crt_dir, ignore_errors=True)
    shutil.rmtree(temp_crt_dir_2, ignore_errors=True)
    shutil.rmtree(temp_storage_dir, ignore_errors=True)
    shutil.rmtree(temp_tracking_dir, ignore_errors=True)


@pytest.fixture(scope="module")
def ca_keypair_path():
    key, crt = tests_util.create_ca_certificate(ca_name='Autossl')
    ca_temp_dir = util.Path(tempfile.mkdtemp())

    ca_crt_path = ca_temp_dir / 'local_ca.crt'
    ca_key_path = ca_temp_dir / 'local_ca.key'

    ca_crt_path.write_bytes(crt)
    ca_key_path.write_bytes(key)

    yield CertificateKeyPair(ca_key_path, ca_crt_path)

    # cleanup temp folders
    shutil.rmtree(str(ca_temp_dir), ignore_errors=True)


@pytest.fixture(scope="module")
def subca_keypair_path(subca_manager, ca_keypair_path):
    storage_path = util.Path(os.environ['AUTOSSL_STORAGE_PATH'])
    key_path = storage_path.joinpath(subca_manager.ssl_blueprint.name + '.key')
    csr_path = storage_path.joinpath(subca_manager.ssl_blueprint.name + '.csr')
    crt_path = storage_path.joinpath(subca_manager.ssl_blueprint.name + '.crt')
    bundle_path = storage_path.joinpath(subca_manager.ssl_blueprint.name + '.bundle')

    # generate sub-CA certificate request and key
    subca_manager.request_renewal(
        force=True,  # disable interactive user input
    )

    # simulate CA signing
    crt_content = tests_util.create_signed_certificate(
        csr_path=csr_path,
        ca_crt_path=ca_keypair_path.crt,
        ca_key_path=ca_keypair_path.key,
        certificate_validity_days=100
    )
    crt_path.write_bytes(crt_content)
    bundle_path.write_bytes(ca_keypair_path.crt.read_bytes() + crt_content)

    yield CertificateKeyPair(key_path, crt_path)


# valid certificate request
def test_subca_ok(tmp_path, subca_manager, subca_keypair_path):
    # check sub-CA certificate
    subca_crt_path, _ = subca_manager.get_and_check_artifacts()
    subca_cert = ssl.SslCertificate().init_from_x509(x509_path=subca_crt_path)
    assert subca_cert.common_name == 'subca.example.com'

    # sign a new certificate with the sub-CA
    _, csr_path = ssl.generate_csr(name='leafcert',
                                   common_name='domain.subca.example.com',
                                   sans=['domain1.subca.example.com', 'domain2.subca.example.com'],
                                   output_path=str(tmp_path))
    crt_content = tests_util.create_signed_certificate(
        csr_path=csr_path,
        ca_crt_path=subca_keypair_path.crt,
        ca_key_path=subca_keypair_path.key
    )
    crt_path = tmp_path / 'leafcert.crt'
    crt_path.write_bytes(crt_content)

    # check trust chain
    bundle_path = os.path.join(os.environ['AUTOSSL_STORAGE_PATH'], subca_manager.ssl_blueprint.name + '.bundle')
    assert os.system('openssl verify -CAfile %s %s' % (bundle_path, crt_path)) == 0


# invalid certificate request, domains are not part of the authorized names
def test_subca_ko(tmp_path, subca_manager, subca_keypair_path):
    # sign a new certificate with the sub-CA
    _, csr_path = ssl.generate_csr(name='invalidcert',
                                   common_name='domain.other.example.com',
                                   sans=['domain.other.example.com'],
                                   output_path=str(tmp_path))
    crt_content = tests_util.create_signed_certificate(
        csr_path=csr_path,
        ca_crt_path=subca_keypair_path.crt,
        ca_key_path=subca_keypair_path.key
    )
    crt_path = tmp_path / 'invalidcert.crt'
    crt_path.write_bytes(crt_content)

    # check trust chain
    bundle_path = os.path.join(os.environ['AUTOSSL_STORAGE_PATH'], subca_manager.ssl_blueprint.name + '.bundle')
    assert os.system('openssl verify -CAfile %s %s' % (bundle_path, crt_path)) != 0



import logging

logger = logging.getLogger(__name__)


class CaManager(object):

    def __init__(self, ca_config, staging=True, storage_api=None, **kwargs):
        """Base instance of interface with CA to deliver signed certificates

        :param ca_config: Certificate Authority configuration instance
        :type ca_config: ssl.CertificateAuthorityConfig
        :param staging: Testing mode. Use staging or test CA instance (when available).
        :type staging: bool
        :param storage_api: storage API instance
        :type storage_api: storage.base.Storage
        """
        self.ca_config = ca_config
        self.staging = staging
        self.storage_api = storage_api

    @property
    def is_automated_renewal_supported(self):
        """Check is current CA supports automated renewal

        :return: True, if this CA implementation supports automated renewal
        :rtype: bool
        """
        return False

    def get_signed_certificate(self, ssl_blueprint=None, csr_path=None, servers_api=None):
        """Get PEM encoded certificate using current Certificate Authority implementation

        :param ssl_blueprint:
        :type ssl_blueprint: ssl.SslBlueprint
        :param csr_path: path to CSR file
        :type csr_path: pathlib.Path
        :param servers_api: list of api instances to each server
        :type servers_api: list(server.base.Server)
        :return: PEM encoded signed certificate as bytes
        :rtype: bytes
        """
        raise NotImplementedError("Method must be overridden in child class")
